@@ -3,7 +3,7 @@
  * from Brian Gladman's implementation.
  *
  * Modifications:
- * - std_defs.h renamed to serpent.h
+ * - std_defs.h renamed to crypto.h
  * - l_key is malloc'ed by set_key rather than being public
  * - reduced to 16 rounds
  *
@@ -40,7 +40,7 @@
 #endif
 
 #include <stdlib.h>
-#include "serpent.h"
+#include "crypto.h"
 
 
 /* Partially optimised Serpent S Box boolean functions derived  */
